@@ -18,7 +18,8 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-import glob
+import fnmatch
+import os
 import os.path
 import shutil
 import sys
@@ -48,10 +49,13 @@ from picard.util.textencoding import (
 )
 from picard.util.filenaming import make_short_filename
 from picard.util.tags import PRESERVED_TAGS
-
+from picard.const import QUERY_LIMIT
+from picard import PICARD_APP_NAME
 
 class File(QtCore.QObject, Item):
 
+    metadata_images_changed = QtCore.pyqtSignal()
+
     UNDEFINED = -1
     PENDING = 0
     NORMAL = 1
@@ -96,16 +100,37 @@ class File(QtCore.QObject, Item):
 
     def load(self, callback):
         thread.run_task(
-            partial(self._load, self.filename),
+            partial(self._load_check, self.filename),
             partial(self._loading_finished, callback),
             priority=1)
 
+    def _load_check(self, filename):
+        # Check that file has not been removed since thread was queued
+        # Don't load if we are stopping.
+        if self.state != File.PENDING:
+            log.debug("File not loaded because it was removed: %r", self.filename)
+            return None
+        if self.tagger.stopping:
+            log.debug("File not loaded because %s is stopping: %r", PICARD_APP_NAME, self.filename)
+            return None
+        return self._load(filename)
+
+    def _load(self, filename):
+        """Load metadata from the file."""
+        raise NotImplementedError
+
     def _loading_finished(self, callback, result=None, error=None):
-        if self.state != self.PENDING:
+        if self.state != File.PENDING or self.tagger.stopping:
             return
         if error is not None:
             self.error = str(error)
             self.state = self.ERROR
+            from picard.formats import supported_extensions
+            file_name, file_extension = os.path.splitext(self.base_filename)
+            if file_extension not in supported_extensions():
+                self.remove()
+                log.error('Unsupported media file %r wrongly loaded. Removing ...',self)
+                return
         else:
             self.error = None
             self.state = self.NORMAL
@@ -140,19 +165,24 @@ class File(QtCore.QObject, Item):
             values = self.orig_metadata.getall(tag)
             if values:
                 saved_metadata[tag] = values
+        deleted_tags = self.metadata.deleted_tags
         self.metadata.copy(metadata)
+        self.metadata.deleted_tags = deleted_tags
         for tag, values in saved_metadata.iteritems():
             self.metadata.set(tag, values)
 
-        self.metadata["acoustid_id"] = acoustid
+        if acoustid:
+            self.metadata["acoustid_id"] = acoustid
+        self.metadata_images_changed.emit()
+
+    def keep_original_images(self):
+        self.metadata.images = self.orig_metadata.images[:]
+        self.update()
+        self.metadata_images_changed.emit()
 
     def has_error(self):
         return self.state == File.ERROR
 
-    def _load(self, filename):
-        """Load metadata from the file."""
-        raise NotImplementedError
-
     def save(self):
         self.set_pending()
         metadata = Metadata()
@@ -165,6 +195,14 @@ class File(QtCore.QObject, Item):
 
     def _save_and_rename(self, old_filename, metadata):
         """Save the metadata."""
+        # Check that file has not been removed since thread was queued
+        # Also don't save if we are stopping.
+        if self.state == File.REMOVED:
+            log.debug("File not saved because it was removed: %r", self.filename)
+            return None
+        if self.tagger.stopping:
+            log.debug("File not saved because %s is stopping: %r", PICARD_APP_NAME, self.filename)
+            return None
         new_filename = old_filename
         if not config.setting["dont_write_tags"]:
             encoded_old_filename = encode_filename(old_filename)
@@ -211,6 +249,11 @@ class File(QtCore.QObject, Item):
             raise OSError
 
     def _saving_finished(self, result=None, error=None):
+        # Handle file removed before save
+        # Result is None if save was skipped
+        if ((self.state == File.REMOVED or self.tagger.stopping)
+            and result is None):
+            return
         old_filename = new_filename = self.filename
         if error is not None:
             self.error = str(error)
@@ -233,17 +276,25 @@ class File(QtCore.QObject, Item):
             for k, v in temp_info.items():
                 self.orig_metadata[k] = v
             self.error = None
-            self.clear_pending()
+            # Force update to ensure file status icon changes immediately after save
+            self.clear_pending(force_update=True)
             self._add_path_to_metadata(self.orig_metadata)
+            self.metadata_images_changed.emit()
 
-        del self.tagger.files[old_filename]
-        self.tagger.files[new_filename] = self
+        if self.state != File.REMOVED:
+            del self.tagger.files[old_filename]
+            self.tagger.files[new_filename] = self
+
+        if self.tagger.stopping:
+            log.debug("Save of %r completed before stopping Picard", self.filename)
 
     def _save(self, filename, metadata):
         """Save the metadata."""
         raise NotImplementedError
 
-    def _script_to_filename(self, format, file_metadata, settings=config.setting):
+    def _script_to_filename(self, format, file_metadata, settings=None):
+        if settings is None:
+            settings = config.setting
         metadata = Metadata()
         if config.setting["clear_existing_tags"]:
             metadata.copy(file_metadata)
@@ -267,8 +318,19 @@ class File(QtCore.QObject, Item):
         filename = filename.replace("\x00", "")
         return filename
 
-    def _make_filename(self, filename, metadata, settings=config.setting):
+    def _fixed_splitext(self, filename):
+        # In case the filename is blank and only has the extension
+        # the real extension is in new_filename and ext is blank
+        new_filename, ext = os.path.splitext(filename)
+        if ext == '' and new_filename.lower() in map(unicode, self.EXTENSIONS):
+            ext = new_filename
+            new_filename = ''
+        return new_filename, ext
+
+    def _make_filename(self, filename, metadata, settings=None):
         """Constructs file name based on metadata and file naming formats."""
+        if settings is None:
+            settings = config.setting
         if settings["move_files"]:
             new_dirname = settings["move_files_to"]
             if not os.path.isabs(new_dirname):
@@ -278,7 +340,7 @@ class File(QtCore.QObject, Item):
         new_filename = os.path.basename(filename)
 
         if settings["rename_files"]:
-            new_filename, ext = os.path.splitext(new_filename)
+            new_filename, ext = self._fixed_splitext(new_filename)
             ext = ext.lower()
             new_filename = new_filename + ext
 
@@ -334,7 +396,12 @@ class File(QtCore.QObject, Item):
         if not metadata.images:
             return
         counters = defaultdict(lambda: 0)
-        for image in metadata.images:
+        images = []
+        if config.setting["caa_save_single_front_image"]:
+            images = metadata.get_single_front_image()
+        if not images:
+            images = metadata.images
+        for image in images:
             image.save(dirname, metadata, counters)
 
     def _move_additional_files(self, old_filename, new_filename):
@@ -343,17 +410,27 @@ class File(QtCore.QObject, Item):
         new_path = encode_filename(os.path.dirname(new_filename))
         patterns = encode_filename(config.setting["move_additional_files_pattern"])
         patterns = filter(bool, [p.strip() for p in patterns.split()])
+        try:
+            names = os.listdir(old_path)
+        except os.error:
+            log.error("Error: {} directory not found".format(old_path))
+            return
+        filtered_names = filter(lambda x: x[0] != '.', names)
         for pattern in patterns:
-            # FIXME glob1 is not documented, maybe we need our own implemention?
-            for old_file in glob.glob1(old_path, pattern):
-                new_file = os.path.join(new_path, old_file)
-                old_file = os.path.join(old_path, old_file)
-                # FIXME we shouldn't do this from a thread!
-                if self.tagger.files.get(decode_filename(old_file)):
-                    log.debug("File loaded in the tagger, not moving %r", old_file)
-                    continue
-                log.debug("Moving %r to %r", old_file, new_file)
-                shutil.move(old_file, new_file)
+            pattern_regex = re.compile(fnmatch.translate(pattern), re.IGNORECASE)
+            file_names = names
+            if pattern[0] != '.':
+                file_names = filtered_names
+            for old_file in file_names:
+                if pattern_regex.match(old_file):
+                    new_file = os.path.join(new_path, old_file)
+                    old_file = os.path.join(old_path, old_file)
+                    # FIXME we shouldn't do this from a thread!
+                    if self.tagger.files.get(decode_filename(old_file)):
+                        log.debug("File loaded in the tagger, not moving %r", old_file)
+                        continue
+                    log.debug("Moving %r to %r", old_file, new_file)
+                    shutil.move(old_file, new_file)
 
     def remove(self, from_parent=True):
         if from_parent and self.parent:
@@ -406,9 +483,13 @@ class File(QtCore.QObject, Item):
                         self.state = File.CHANGED
                     break
         else:
-            self.similarity = 1.0
-            if self.state in (File.CHANGED, File.NORMAL):
-                self.state = File.NORMAL
+            if (self.metadata.images and
+               self.orig_metadata.images != self.metadata.images):
+                self.state = File.CHANGED
+            else:
+                self.similarity = 1.0
+                if self.state in (File.CHANGED, File.NORMAL):
+                    self.state = File.NORMAL
         if signal:
             log.debug("Updating file %r", self)
             if self.item:
@@ -551,6 +632,7 @@ class File(QtCore.QObject, Item):
         )
         self.clear_lookup_task()
         metadata = self.metadata
+        self.set_pending()
         self.lookup_task = self.tagger.xmlws.find_tracks(partial(self._lookup_finished, 'metadata'),
             track=metadata['title'],
             artist=metadata['artist'],
@@ -559,7 +641,7 @@ class File(QtCore.QObject, Item):
             tracks=metadata['totaltracks'],
             qdur=str(metadata.length / 2000),
             isrc=metadata['isrc'],
-            limit=25)
+            limit=QUERY_LIMIT)
 
     def clear_lookup_task(self):
         if self.lookup_task:
@@ -571,10 +653,12 @@ class File(QtCore.QObject, Item):
             self.state = File.PENDING
             self.update()
 
-    def clear_pending(self):
+    def clear_pending(self, force_update=False):
         if self.state == File.PENDING:
             self.state = File.NORMAL
             self.update()
+        elif force_update:
+            self.update()
 
     def iterfiles(self, save=False):
         yield self
